import asyncio
import os

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from sqlalchemy.orm import Session

os.environ["ARGOS_APP_ENV"] = "test"


@pytest.fixture
def db() -> Session:
    from argos.server import models

    app = _create_app()
    models.Base.metadata.create_all(bind=app.state.engine)
    yield app.state.SessionLocal()
    models.Base.metadata.drop_all(bind=app.state.engine)


@pytest.fixture
def app() -> FastAPI:
    from argos.server import models

    app = _create_app()
    models.Base.metadata.create_all(bind=app.state.engine)
    yield app
    models.Base.metadata.drop_all(bind=app.state.engine)


@pytest.fixture
def authorized_client(app):
    with TestClient(app) as client:
        token = app.state.config.service.secrets[0]
        client.headers = {"Authorization": f"Bearer {token}"}
        yield client


def _create_app() -> FastAPI:
    from argos.server.main import (  # local import for testing purpose
        get_application,
        setup_database,
        connect_to_db,
    )

    app = get_application()
    # Hardcode the database url and the yaml file for testing purpose
    # Otherwise, the app will try to read the .env file or the environment variables
    app.state.settings.database_url = "sqlite:////tmp/test-argos.db"
    app.state.settings.yaml_file = "tests/config.yaml"

    setup_database(app)
    asyncio.run(connect_to_db(app))
    return app



"""Web interface for humans"""
from collections import defaultdict
from functools import cmp_to_key
from os import path
from typing import Annotated
from urllib.parse import urlparse

from fastapi import APIRouter, Cookie, Depends, Form, Request, status
from fastapi.responses import RedirectResponse
from fastapi.templating import Jinja2Templates
from sqlalchemy import func
from sqlalchemy.orm import Session

from argos.schemas import Config
from argos.server import queries
from argos.server.models import Result, Task
from argos.server.routes.dependencies import get_config, get_db

route = APIRouter()

current_dir = path.dirname(__file__)
templates = Jinja2Templates(directory=path.join(current_dir, "../templates"))
SEVERITY_LEVELS = {"ok": 1, "warning": 2, "critical": 3, "unknown": 4}


@route.get("/")
async def get_severity_counts_view(
    request: Request,
    db: Session = Depends(get_db),
    auto_refresh_enabled: Annotated[bool, Cookie()] = False,
    auto_refresh_seconds: Annotated[int, Cookie()] = 15,
):
    """Shows the number of results per severity"""
    counts_dict = await queries.get_severity_counts(db)

    agents = db.query(Result.agent_id).distinct().all()

    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "counts_dict": counts_dict,
            "agents": agents,
            "auto_refresh_enabled": auto_refresh_enabled,
            "auto_refresh_seconds": auto_refresh_seconds,
        },
    )


@route.get("/domains")
async def get_domains_view(request: Request, db: Session = Depends(get_db)):
    """Show all tasks and their current state"""
    tasks = db.query(Task).all()

    domains_severities = defaultdict(list)
    domains_last_checks = defaultdict(list)

    for task in tasks:
        domain = urlparse(task.url).netloc
        domains_severities[domain].append(task.severity)
        if task.last_severity_update is not None:
            domains_last_checks[domain] = task.last_severity_update
        else:
            domains_last_checks[domain] = "Waiting to be checked"

    def _max_severity(severities):
        return max(severities, key=SEVERITY_LEVELS.get)

    def _cmp_domains(a, b):
        if SEVERITY_LEVELS[a[1]] < SEVERITY_LEVELS[b[1]]:
            return 1
        if SEVERITY_LEVELS[a[1]] > SEVERITY_LEVELS[b[1]]:
            return -1
        if a[0] > b[0]:
            return 1
        if a[0] < b[0]:
            return -1
        return 0

    domains = [(key, _max_severity(value)) for key, value in domains_severities.items()]
    domains.sort(key=cmp_to_key(_cmp_domains))

    agents = db.query(Result.agent_id).distinct().all()

    return templates.TemplateResponse(
        "domains.html",
        {
            "request": request,
            "domains": domains,
            "last_checks": domains_last_checks,
            "total_task_count": len(tasks),
            "agents": agents,
        },
    )


@route.get("/domain/{domain}")
async def get_domain_tasks_view(
    request: Request, domain: str, db: Session = Depends(get_db)
):
    """Show all tasks attached to a domain"""
    tasks = db.query(Task).filter(Task.domain.contains(f"//{domain}")).all()
    return templates.TemplateResponse(
        "domain.html", {"request": request, "domain": domain, "tasks": tasks}
    )


@route.get("/result/{result_id}")
async def get_result_view(
    request: Request, result_id: int, db: Session = Depends(get_db)
):
    """Show the details of a result"""
    result = db.query(Result).get(result_id)
    return templates.TemplateResponse(
        "result.html", {"request": request, "result": result}
    )


@route.get("/task/{task_id}/results")
async def get_task_results_view(
    request: Request,
    task_id: int,
    db: Session = Depends(get_db),
    config: Config = Depends(get_config),
):
    """Show history of a task’s results"""
    results = (
        db.query(Result)
        .filter(Result.task_id == task_id)
        .order_by(Result.submitted_at.desc())  # type: ignore[attr-defined]
        .all()
    )
    task = db.query(Task).get(task_id)
    description = task.get_check().get_description(config)
    return templates.TemplateResponse(
        "results.html",
        {
            "request": request,
            "results": results,
            "task": task,
            "description": description,
        },
    )


@route.get("/agents")
async def get_agents_view(request: Request, db: Session = Depends(get_db)):
    """Show argos agents and the last time the server saw them"""
    last_seen = (
        db.query(Result.agent_id, func.max(Result.submitted_at).label("submitted_at"))
        .group_by(Result.agent_id)
        .all()
    )

    return templates.TemplateResponse(
        "agents.html", {"request": request, "last_seen": last_seen}
    )


@route.post("/refresh")
async def set_refresh_cookies_view(
    request: Request,
    auto_refresh_enabled: Annotated[bool, Form()] = False,
    auto_refresh_seconds: Annotated[int, Form()] = 15,
):
    response = RedirectResponse(
        request.url_for("get_severity_counts_view"),
        status_code=status.HTTP_303_SEE_OTHER,
    )
    response.set_cookie(key="auto_refresh_enabled", value=auto_refresh_enabled)
    response.set_cookie(key="auto_refresh_seconds", value=int(auto_refresh_seconds))
    return response



import os
import sys

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError
from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker

from argos.logging import logger
from argos.server import models, routes
from argos.server.settings import get_app_settings, read_yaml_config


def get_application() -> FastAPI:
    """Spawn Argos FastAPI server"""
    settings = get_app_settings()
    appli = FastAPI()

    config = read_config(appli, settings)

    # Settings is the pydantic settings object
    # Config is the argos config object (built from yaml)
    appli.state.config = config
    appli.state.settings = settings

    appli.add_event_handler(
        "startup",
        create_start_app_handler(appli),
    )
    appli.add_event_handler(
        "shutdown",
        create_stop_app_handler(appli),
    )
    appli.include_router(routes.api, prefix="/api")
    appli.include_router(routes.views)

    static_dir = os.path.join(os.path.dirname(__file__), "static")

    appli.mount("/static", StaticFiles(directory=static_dir), name="static")
    return appli


def create_start_app_handler(appli):
    """Warmup the server:
    setup database connection
    """

    async def _get_db():
        setup_database(appli)

        return await connect_to_db(appli)

    return _get_db


async def connect_to_db(appli):
    appli.state.db = appli.state.SessionLocal()
    return appli.state.db


def create_stop_app_handler(appli):
    """Gracefully shutdown the server:
    close database connection.
    """

    async def stop_app():
        appli.state.db.close()

    return stop_app


def read_config(appli, settings):
    try:
        config = read_yaml_config(settings.yaml_file)
        appli.state.config = config
        return config
    except ValidationError as err:
        logger.error("Errors where found while reading configuration:")
        for error in err.errors():
            logger.error("%s is %s", error["loc"], error["type"])
        sys.exit(1)


def setup_database(appli):
    settings = appli.state.settings
    # For sqlite, we need to add connect_args={"check_same_thread": False}
    logger.debug("Using database URL %s", settings.database_url)
    if settings.database_url.startswith("sqlite:////tmp"):
        logger.warning("Using sqlite in /tmp is not recommended for production")

    extra_settings = {}
    if settings.db_pool_size:
        extra_settings.setdefault("pool_size", settings.db_pool_size)

    if settings.db_max_overflow:
        extra_settings.setdefault("max_overflow", settings.db_max_overflow)

    engine = create_engine(settings.database_url, **extra_settings)

    def _fk_pragma_on_connect(dbapi_con, con_record):
        dbapi_con.execute("pragma foreign_keys=ON")

    if settings.database_url.startswith("sqlite:////"):
        event.listen(engine, "connect", _fk_pragma_on_connect)

    appli.state.SessionLocal = sessionmaker(
        autocommit=False, autoflush=False, bind=engine
    )
    appli.state.engine = engine
    models.Base.metadata.create_all(bind=engine)


app = get_application()



"""Pydantic schemas for server"""
import os
from functools import lru_cache
from os import environ
from typing import Optional, Union

import yaml
from pydantic_settings import BaseSettings, SettingsConfigDict
from yamlinclude import YamlIncludeConstructor

from argos.schemas.config import Config


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="argos_", env_file=".env")
    app_env: str
    database_url: str
    yaml_file: str
    db_pool_size: Optional[int]
    db_max_overflow: Optional[int]


class DevSettings(Settings):
    """Settings for dev environment.

    Uses config.yaml as config file.
    Uses a SQLite database."""

    app_env: str = "dev"
    yaml_file: str = "config.yaml"
    db_pool_size: Optional[int] = None
    db_max_overflow: Optional[int] = None
    database_url: str = "sqlite:////tmp/argos.db"


class TestSettings(Settings):
    """Settings for test environment.

    Uses tests/config.yaml as config file.
    Uses a SQLite database."""

    app_env: str = "test"
    yaml_file: str = "tests/config.yaml"
    database_url: str = "sqlite:////tmp/test-argos.db"
    db_pool_size: Optional[int] = None
    db_max_overflow: Optional[int] = None


class ProdSettings(Settings):
    """Settings for prod environment."""

    app_env: str = "prod"
    db_pool_size: Optional[int] = 10
    db_max_overflow: Optional[int] = 20


environments = {
    "dev": DevSettings,
    "prod": ProdSettings,
    "test": TestSettings,
}


@lru_cache()
def get_app_settings() -> Union[None, Settings]:
    """Load settings depending on the environment"""
    app_env = environ.get("ARGOS_APP_ENV", "dev")
    settings = environments.get(app_env)
    if settings is not None:
        return settings()
    return None


def read_yaml_config(filename):
    parsed = _load_yaml(filename)
    return Config(**parsed)


def _load_yaml(filename):
    base_dir = os.path.dirname(filename)
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.FullLoader, base_dir=base_dir
    )

    with open(filename, "r", encoding="utf-8") as stream:
        return yaml.load(stream, Loader=yaml.FullLoader)



import asyncio
import os
from functools import wraps
from uuid import uuid4

import click
import uvicorn
from alembic import command
from alembic.config import Config

from argos import logging
from argos import VERSION
from argos.agent import ArgosAgent


async def get_db():
    from argos.server.main import connect_to_db, get_application, setup_database

    app = get_application()
    setup_database(app)
    return await connect_to_db(app)


def coroutine(f):
    """Decorator to enable async functions in click"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def validate_config_access(ctx, param, value):
    if os.path.isfile(value) and os.access(value, os.R_OK):
        return value

    if os.path.isfile(value):
        raise click.BadParameter(f"the file {value} is not readabale.")

    raise click.BadParameter(f"the file {value} does not exists or is not reachable.")


@click.group()
def cli():
    pass


@cli.group()
def server():
    pass


@cli.command()
def version():
    click.echo(VERSION)


@cli.command()
@click.argument("server_url", envvar="ARGOS_AGENT_SERVER_URL")
@click.argument("auth", envvar="ARGOS_AGENT_TOKEN")
@click.option(
    "--max-tasks",
    default=10,
    help="Number of concurrent tasks this agent can run",
)
@click.option(
    "--wait-time",
    default=10,
    help="Waiting time between two polls on the server (seconds)",
)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(logging.LOG_LEVELS, case_sensitive=False),
)
def agent(server_url, auth, max_tasks, wait_time, log_level):
    """Get and run tasks to the provided server. Will wait for new tasks.

    Usage: argos agent https://argos.example.org "auth-token-here"

    Alternatively, you can use the following environment variables to avoid passing
    arguments to the agent on the command line:

      \b
      ARGOS_AGENT_SERVER_URL=https://argos.example.org
      ARGOS_AGENT_TOKEN=auth-token-here
    """
    click.echo("Starting argos agent. Will retry forever.")
    from argos.logging import logger

    logger.setLevel(log_level)
    agent_ = ArgosAgent(server_url, auth, max_tasks, wait_time)
    asyncio.run(agent_.run())


@server.command()
@click.option("--host", default="127.0.0.1", help="Host to bind")
@click.option("--port", default=8000, type=int, help="Port to bind")
@click.option(
    "--config",
    default="config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@click.option("--reload", is_flag=True, help="Enable hot reloading")
def start(host, port, config, reload):
    """Starts the server (use only for testing or development!)

    See https://framasoft.frama.io/framaspace/argos/deployment/systemd.html#server
    for advices on how to start the server for production.
    """
    os.environ["ARGOS_YAML_FILE"] = config
    uvicorn.run("argos.server:app", host=host, port=port, reload=reload)


def validate_max_lock_seconds(ctx, param, value):
    if value <= 60:
        raise click.BadParameter("Should be strictly higher than 60")
    return value


def validate_max_results(ctx, param, value):
    if value <= 0:
        raise click.BadParameter("Should be a positive integer")
    return value


@server.command()
@click.option(
    "--max-results",
    default=100,
    help="Number of results per task to keep",
    callback=validate_max_results,
)
@click.option(
    "--max-lock-seconds",
    default=100,
    help=(
        "The number of seconds after which a lock is "
        "considered stale, must be higher than 60 "
        "(the checks have a timeout value of 60 seconds)"
    ),
    callback=validate_max_lock_seconds,
)
@click.option(
    "--config",
    default="config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def cleandb(max_results, max_lock_seconds, config):
    """Clean the database (to run routinely)

    \b
    - Removes old results from the database.
    - Removes locks from tasks that have been locked for too long.
    """
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries

    db = await get_db()
    removed = await queries.remove_old_results(db, max_results)
    updated = await queries.release_old_locks(db, max_lock_seconds)

    click.echo(f"{removed} results removed")
    click.echo(f"{updated} locks released")


@server.command(short_help="Load or reload tasks’ configuration")
@click.option(
    "--config",
    default="config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def reload_config(config):
    """Read tasks’ configuration and add/delete tasks in database if needed"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server import queries
    from argos.server.main import get_application, read_config
    from argos.server.settings import get_app_settings

    appli = get_application()
    settings = get_app_settings()
    config = read_config(appli, settings)

    db = await get_db()
    changed = await queries.update_from_config(db, config)

    click.echo(f"{changed['added']} tasks added")
    click.echo(f"{changed['vanished']} tasks deleted")


@server.command()
@click.option(
    "--config",
    default="config.yaml",
    help="Path of the configuration file. "
    "If ARGOS_YAML_FILE environment variable is set, its value will be used instead.",
    envvar="ARGOS_YAML_FILE",
    callback=validate_config_access,
)
@coroutine
async def migrate(config):
    """Run database migrations"""
    # It’s mandatory to do it before the imports
    os.environ["ARGOS_YAML_FILE"] = config

    # The imports are made here otherwise the agent will need server configuration files.
    from argos.server.settings import get_app_settings

    settings = get_app_settings()

    current_dir = os.path.dirname(__file__)
    alembic_cfg = Config(os.path.join(current_dir, "server/migrations/alembic.ini"))
    alembic_cfg.set_main_option("sqlalchemy.url", settings.database_url)
    command.upgrade(alembic_cfg, "head")


@server.command(short_help="Generate a token for agents")
@coroutine
async def generate_token():
    """Generate a token, which can be used as an agent’s authentication token.

    It’s actually an UUID
    """
    click.echo(uuid4())


if __name__ == "__main__":
    cli()
